/* string_transmission_protocoll.c
 *
 * Copyright (c) 2010 Stefan Beller.  Use may be in
 * whole or in part in accordance to the General Public License (GPL).
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
 
#include "string_transport_protocoll.h"

void __stp_init_connectionstruct(stp_connection stp,int socket){

}

#define STP_CONNECT_FAIL(msg) {perror(msg);return NULL;}
stp_connection stp_connect(char* host, int port){
	int sockfd;
	struct sockaddr_in dest;
	struct hostent *hp;
	stp_connection ret;

	/*---Create socket for streaming---*/
	if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) STP_CONNECT_FAIL("stp_connect: Socket creation failed.");

	/*---Initialize server address/port struct---*/
	dest.sin_family = AF_INET;
	dest.sin_port = htons(port);	
    if((hp = gethostbyname(host)) == NULL) STP_CONNECT_FAIL("stp_connect: gethostbyname failed.");
    memmove(&dest.sin_addr,hp->h_addr, hp->h_length);  

	/*---Connect to server---*/
	if ( connect(sockfd, (struct sockaddr *)&dest, sizeof(dest)) != 0 ) STP_CONNECT_FAIL("stp_connect: connect failed.")

	/*---get memory to return a struct---*/
	if ( (ret=(stp_connection)malloc(sizeof(stp_connectionstruct)))==NULL) STP_CONNECT_FAIL("stp_connect: malloc failed.")

	ret->waitingline = NULL;
	ret->maxSize = 0;
	ret->socket = sockfd;
	ret->offset = 0;
	
	if( pthread_mutex_init( &ret->rmutex, 0 ) != 0)
			STP_CONNECT_FAIL("stp_connect: pthread_mutex_init")
			
	if( pthread_mutex_init( &ret->smutex, 0 ) !=0 )
			STP_CONNECT_FAIL("stp_connect: pthread_mutex_init")
			
	return ret;
}

#define __STP_ACCEPT_CONNECTIONS_FAIL(msg) {perror(msg);p->callback(NULL);}
void* __stp_accept_connections(void* v){
	int sd;
    int client;
    struct sockaddr_in addr;
	unsigned int addr_size;
	stp_connection stp;
	stp_accepter p;
	
	p = (stp_accepter) v;	
	
	addr.sin_family = AF_INET;
    addr.sin_port = htons( p->port );
    addr.sin_addr.s_addr = INADDR_ANY; 
	
    if (   (sd = socket(AF_INET, SOCK_STREAM , 0)) < 0   ) 
    		__STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: socket");
   
    if (   bind(sd, (struct sockaddr*)&addr, sizeof(addr)) != 0    ) 
    		__STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: bind.");
   
    if (   listen(sd, SOMAXCONN) != 0    ) 
    		__STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: listen");

    while (p->running){

			addr_size = sizeof(addr);
			
			client = accept(sd, (struct sockaddr*)&addr, &addr_size);
			if(p->running){
				if( (stp = (stp_connection)malloc(sizeof(stp_connectionstruct))) == NULL) 
						__STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: malloc");
				stp->waitingline = NULL;
				stp->maxSize = 0;
				stp->socket = client;
				stp->offset = 0;
				if( pthread_mutex_init( &stp->rmutex, 0 ) !=0)
						__STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: pthread_mutex_init");
				if( pthread_mutex_init( &stp->smutex, 0 ) !=0)
						__STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: pthread_mutex_init");
				p->callback(stp);
			}
    }
    close(client);
    /*do not shutdown s, because this is done by callback*/
    /*do not free the parameter we got by (void* v), because this is done by stp_stop_accepting */
    return v;
}

#define STP_ACCEPT_CONNECTIONS_FAIL(msg) {perror(msg);return NULL;}
stp_accepter stp_accept_connections(int port, void callback(stp_connection new_connection)){
	stp_accepter p;
	
	if( (p = (stp_accepter) malloc(sizeof(stp_acceptingstruct))) == NULL)
			STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: malloc");
	
	if( (p->accepting_thread = (pthread_t*) malloc(sizeof(pthread_t))) == NULL)
			STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: malloc");
			
	if( (p->accepting_thread_attr = (pthread_attr_t*) malloc(sizeof(pthread_attr_t))) == NULL)
			STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: malloc");
	
	p->port = port;
	p->callback = callback;
	p->running = 1;
	
	if (pthread_attr_init(p->accepting_thread_attr)!=0)
			STP_ACCEPT_CONNECTIONS_FAIL("stp_accept_connections: pthread_attr_init");
	
	if ( pthread_create(p->accepting_thread, p->accepting_thread_attr, __stp_accept_connections, p) != 0 ) 
			STP_ACCEPT_CONNECTIONS_FAIL("pthread_create"); 
			
	return p;
}

void stp_stop_accepting(stp_accepter accepter){
	stp_connection s;
	char* host = "localhost";
	accepter->running=0; 
	
	/*still accepting, so send a dummy connection, which can be closed directly*/
	if( (s = stp_connect(host,accepter->port)) == NULL)
		{perror("stp_stop_accepting: connect");fflush(stderr);exit(0);}
	sleep(1);
	stp_shutdown(s);

	/*now wait until it is finished and free pthread ressources.*/
	pthread_join( *(accepter->accepting_thread), 0 );
	pthread_attr_destroy(accepter->accepting_thread_attr);
	pthread_detach( *(accepter->accepting_thread) );

	/*now free all stuff, we malloc'd*/
	free(accepter->accepting_thread_attr);
	free(accepter->accepting_thread);
	free(accepter);
}


#define STP_RECV_FAIL(msg) {perror(msg);return NULL;}
char* stp_recv(stp_connection stp ){
	int bytes_read, bytes_read_sum;
	char buffer[1];
	char* ret;
	ret = NULL;
	
	stp_length_t length;
	
	bytes_read_sum=0;
	bytes_read=0;
	pthread_mutex_lock(&stp->rmutex);
	
	while( bytes_read_sum < sizeof(STP_MAXLENGTH_DATATYPE) ){
			
		bytes_read = recv(stp->socket, &(length.length_str[bytes_read_sum]), \
					 sizeof(STP_MAXLENGTH_DATATYPE)-bytes_read_sum, 0);
					 
		if(bytes_read>0)
			bytes_read_sum += bytes_read;
		else if(bytes_read == 0)
			usleep(STP_RECV_SLEEPTIME);	
		else
			STP_RECV_FAIL("stp_recv: recv failed: 1:");
	}	
	
	/*
	 * check if length is zero, then the othersided socket was shutdown.
	 * Even strings like "" have a length of 1, the '\0' 
	 */
	if( length.length_int == 0 ) /* ret is still NULL*/ goto stp_recv_exit;
	
	if( (stp->waitingline = (char*)malloc(length.length_int*sizeof(char)) ) == NULL)STP_RECV_FAIL("stp_recv: malloc failed: 2");
	
	stp->offset = 0;
	stp->maxSize = length.length_int;
	while(!ret){	
		bytes_read = recv(stp->socket, buffer, 1, 0);
		if( bytes_read == 1 ){
			if( stp->offset >= stp->maxSize ){
				stp->maxSize*=2;
				if( (stp->waitingline = realloc(stp->waitingline,stp->maxSize)) == NULL)STP_RECV_FAIL("stp_recv: realloc failed: 3");
				fprintf(stderr, "stp_recv: length invalid, reallocate memory\n");
			}
			stp->waitingline[stp->offset] = buffer[0];
			stp->offset++;
							
			if( buffer[0] == 0 ){
				ret = stp->waitingline;
				stp->waitingline=NULL;
			}
			
		}else if(bytes_read == 0){
			/*nothing has come, so wait*/
			usleep(STP_RECV_SLEEPTIME);
		}else
			STP_RECV_FAIL("stp_recv: recv failed: 4");
	}
	
stp_recv_exit:	
	pthread_mutex_unlock(&stp->rmutex);
	return ret;
}

int stp_send(stp_connection stp, char* line){
	int ret;
	int bytes_sent;
	int bytes_sent_sum;
	stp_length_t length;
	
	ret=0;
	bytes_sent=0;
	bytes_sent_sum=0;
	
	length.length_int = strlen(line)+1;
	
	pthread_mutex_lock(&stp->smutex);//get mutex before touching the socket!

	while( bytes_sent_sum < sizeof(STP_MAXLENGTH_DATATYPE) ){	
		bytes_sent = send(stp->socket,&(length.length_str[bytes_sent_sum]), \
					 sizeof(STP_MAXLENGTH_DATATYPE) - bytes_sent_sum,0);
		if(bytes_sent>=0){
				bytes_sent_sum += bytes_sent;
				break;
		}else{
				/*an error occured, return 1 to indicate*/
				perror("stp_send");
				ret = 1;
				goto stp_send_exit;
				break;
		}
	}

	while(length.length_int>0){

		bytes_sent = send(stp->socket,line,length.length_int,0);	

		if(errno){
			/*an error occured, return 1 to indicate*/
			ret = 1;
			goto stp_send_exit;
		}

		switch(bytes_sent){
			default:
			case 1:
				/*set line and length to the remaining string, which is still not send*/
				line+=bytes_sent;
				length.length_int-=bytes_sent;
				break;
			case 0:
				/* nothing sent, but no error, so continue*/
				break;
			case -1:
				/*an error occured, return 1 to indicate*/
				perror("stp_send");
				ret = 1;
				goto stp_send_exit;
				break;
		}
	}
stp_send_exit:
	pthread_mutex_unlock(&stp->smutex);
	return ret;
}
 

void stp_shutdown(stp_connection stp ){
	/* to shutdown a socket, send a to other socket length=0, so that socket can shutdown as well. */
	int bytes_sent;
	int bytes_sent_sum;
	stp_length_t length;	
	bytes_sent=0;
	bytes_sent_sum=0;
	length.length_int = 0;
	while( bytes_sent_sum < sizeof(STP_MAXLENGTH_DATATYPE) ){	
		bytes_sent = send(stp->socket,&(length.length_str[bytes_sent_sum]), \
					 sizeof(STP_MAXLENGTH_DATATYPE) - bytes_sent_sum,0);
		if(bytes_sent>=0){
				bytes_sent_sum += bytes_sent;
		}else{
				/* an error occured, while trying to send;
				 * because of this error, the contact to the other socket 
				 * should be considered completely broken.
				 * set bytes_sent_sum, so that the while-condition terminates
				 */
				bytes_sent_sum = sizeof(STP_MAXLENGTH_DATATYPE);
		}
	}
	pthread_mutex_destroy( &stp->smutex );
	pthread_mutex_destroy( &stp->rmutex );
	close(stp->socket);
	free(stp);
}

/*EOF*/
